// alloc 1024 memory blocks and test for errors
//   free good memory blocks
//   do not free bad memory blocks and run (+alloc bad blocks) forever
// if some bits are bad, 0.1% of memory will be occupied and rest can be used
//
// gcc -Wall -O2 -o mem_badram mem_badram.c
// ./mem_badram 31000  # argument1 is physical memory in MB
//
// 2013-03-22 improved version
// 2011-07-06 cosmetic changes for output (more info)

#include <stdlib.h>
#include <stdio.h>

typedef unsigned long int ULong;

// analyse memory given by pointer and length in sizeoflong
int chkmem(ULong *a, ULong num, ULong rnd) {
  ULong jj=0, i, j, lrnd=rnd;
  for (j=1;j<16; j++){
    lrnd=~lrnd;  /* inverse bitpattern between tests */
    for (i=0; i<num; i++) a[i]=rnd;
    for (i=0; i<num; i++) if (a[i]!=rnd) {
      printf("ERROR bad data at a[%08lx]=%08lx != %08lx diff=%08lx (j=%lu)\n",
        i,(ULong)a[i],rnd,(a[i]^rnd),j); 
        jj=i; break;
    }
    if (jj) break;
  }
  return jj;
}

int main(int argn, char *argv[]){
  ULong rnd, rnd1, rnd2, ii, i, j, *pa[1024], bad=0, num, jj=0, nb=0, mb=1024;
  int sl=sizeof(ULong);            /* sizeLong should be 8 on 64bit or 4 on 32bit */
  if (argn>1) mb=atoi(argv[1]);    /* memory in MB */
  num = mb*(1024l/sizeof(ULong));  /* memory in kB/sizeLong */
  printf(" malloc %lu MiB (Mi=2^20) long= %d B = %d bit\n", mb, sl, sl*8);
  // split into 1024 blocks of size mb*1024
  printf(" malloc 1024 blocks to %lu kiB = %lu * long (ki=2^10)\n", mb, num);
  for (i=0;i<1024;i++) {
    pa[i] = (ULong *) malloc(mb*1024l); /* alloc mb*kB (memoryMB/1024) */
    if ((i&127)==0) printf(" a=%012lx = %3lu GB blk %4lu (alloc+touch)\n",
                           (ULong)pa[i], ((ULong)pa[i])>>30, i);
    if (!pa[i]) break;
    for (j=0; j<num; j+=512) pa[i][j]=0; // touch every 512 B!
  } nb=i;
  if (nb!=1024)
    printf(" WARNING: only %lu blocks = %lu kiB allocated\n", nb+1, (nb+1)*mb);
  if (!nb) exit(1);
  for (ii=0;ii<100;ii++) {
   rnd1=rand();
   rnd2=rand();
   rnd=(rnd1<<32) | rnd2;
   if ((ii&7)==0) rnd=0;
   // printf(" rnd %016lx %016lx %016lx\n", rnd1, rnd2, rnd);
   for (i=0;i<nb;i++) {
    if ((i&127)==0) printf(" a=%012lx blk %4lu (ii=%lu)*8 %016lx\n",
                           (ULong)pa[i], i, ii, rnd);
    if (pa[i]) jj = chkmem(pa[i],num,rnd); // 0=OK
    if (jj) { bad++; pa[i]=0; } // mark as bad
   }
  }
  for (i=0;i<nb;i++) if (pa[i]) free(pa[i]); // free good blocks
  if (bad) for (;;) system("sleep 600"); // sleep forever
  return 0;
}
